/*  Huggy Bear mini sumo code. 
 *   
 * Much of this is ported from DeltaForce Mini Sumo code from 2001.
 * Source provided for informational purposes only with no warranty what so ever.
 *  
 */
#include <Wire.h>
#include <VL53L0X.h>
#include <Streaming.h>
#include <ADC.h>
#include <biquad.h>
//#include <oscillator.h>
#include <FirstOrderFilter.h>
#include <StateVariableFilter.h>

//dir_control constants.
   
   #define FWD 1
   #define FWD_LEFT 2
   #define FWD_RIGHT 3
   #define FWD_HARD_LEFT 4
   #define FWD_HARD_RIGHT 5
   #define FWD_MAX 6
   
   #define SPIN_LEFT 10
   #define SPIN_RIGHT 11
   #define STOP 12
   #define BRAKE 13

   #define REV 16
   #define REV_LEFT 17
   #define REV_RIGHT 18
   #define REV_HARD_LEFT 19
   #define REV_HARD_RIGHT 20
   #define REV_MAX 21
   

enum roller {ROLLERS_OFF,ROLLERS_ON, ROLLERS_REV};

enum op_dir {OD_NONE,OD_LEFT,OD_RIGHT,OD_CENTER, OD_CENTER_NEAR};
int  opponentLocation;

int cruse_output, chase_output ,  maneuver_output, deadlock_output;
int check_battery_output, continue_output, arbitrate_output;
int m_action, m_timer, m_border, m_opponent, m_priority;

const int lMotor = 4;
const int rMotor = 3;
const int enableMotor = 12;
const int lDoppler = A1;
const int rDoppler = A0;
const int POT = A6;
const int LED = 13;
const int LED2 = 7;
const int DAC = A14;    //teensy 3.2 hardware DAC
const int lLine = 11;
const int rLine = 10;
const int rollerPWM = 5;
const int rollerDIR = 6;
const int TEST = 8;
const int fs = 4800;   //sample rate

VL53L0X distance_sensor;     //Time of flight optical distance sensor

IntervalTimer sampleTimer;
ADC *adc = new ADC();

volatile int32_t lDC,rDC,aim;

FirstOrderFilter lpR1 = FirstOrderFilter( FirstOrderFilter::LOWPASS,fs, 50.0);              //50Hz lp for DC blocker
StateVariableFilter lpR2 = StateVariableFilter(StateVariableFilter::LOWPASS,fs,10, 0.707);  //10Hz lp Ripple filter

FirstOrderFilter lpL1 = FirstOrderFilter( FirstOrderFilter::LOWPASS,fs, 50.0);               //50Hz lp for DC blocker
StateVariableFilter lpL2 = StateVariableFilter(StateVariableFilter::LOWPASS,fs,10, 0.707);   //10Hz lp Ripple filter

// 20 hz low pass filters for white border line detectors
StateVariableFilter lpLeftLine = StateVariableFilter(StateVariableFilter::LOWPASS,fs,20, 0.707);
StateVariableFilter lpRightLine = StateVariableFilter(StateVariableFilter::LOWPASS,fs,20, 0.707);



volatile int32_t rOpponent, lOpponent, pot ;
int spd;
uint64_t ptime;

int32_t leftLineFiltered,rightLineFiltered;

int l_speed,r_speed;

volatile int  prescale, timer, runTimer, counter1,divide10,divide100;
volatile int  timer2, startup_timer, motor_timer, tilt_dir, escape_timer, distance_sensor_timer;
volatile int  tilt_memory, border_memory, pushback_memory, cruse_memory;
volatile int leftBorder,rightBorder;

int arb_code,dir_ctl;
int flags;
unsigned int distance_mm;

//Bits in flags variable
#define tick100 0
#define direction 1
#define no_IR 2
#define mot_stp_req 3
#define mot_stp_ack 4
#define spin_right 5
#define attack 6
#define startup 7


//-------------------------------------------------------------------------------------------------

void setup() {


 Serial.begin(9600);

//Setup I2C
 Wire.begin();

//Setup VL53L0X time of flight optical distance sensor
 distance_sensor.init();
 distance_sensor.setTimeout(100);
 distance_sensor.startContinuous();
 
 analogWriteResolution(8);
 
 sampleTimer.begin(procSample,1000000/fs);  //Setup periodic interrupt 4800 Hz

 adc->setResolution(12,ADC_0);
 adc->setConversionSpeed(ADC_VERY_HIGH_SPEED,ADC_0);
 adc->setSamplingSpeed(ADC_VERY_HIGH_SPEED,ADC_0);
 adc->setAveraging(0,ADC_0);

 adc->setResolution(12,ADC_1);
 adc->setConversionSpeed(ADC_VERY_HIGH_SPEED,ADC_1);
 adc->setSamplingSpeed(ADC_VERY_HIGH_SPEED,ADC_1);
 adc->setAveraging(0,ADC_1);

 pinMode(TEST,OUTPUT);  //For test and debug
 digitalWrite(TEST,HIGH);

 pinMode(rMotor,OUTPUT);    //Right motor lockec antiphase PWM
 pinMode(lMotor,OUTPUT);    //Left motor locked antiphase PWM
 pinMode(enableMotor,OUTPUT);

 pinMode(lLine,INPUT);      //Left line detector, high true
 pinMode(rLine,INPUT);      //Right line detector, high true

 pinMode(LED,OUTPUT);
 pinMode(LED2,OUTPUT);
 
 pinMode(rollerPWM,OUTPUT);
 pinMode(rollerDIR,OUTPUT);
 rollers(ROLLERS_OFF);
  
 analogWriteFrequency(4,20000);   //20khz pwm freq
 analogWrite(lMotor,128);         //Motors stopped
 analogWrite(rMotor,128);
 digitalWrite(enableMotor,LOW);    //disable motors

  prescale = 220;

  Serial.print("Start");

 delay(4600);                         //wait 4.6 secs plus the .4 sec startup delay in teensy 3.2
  
  

  ptime = millis() + 500;
 

  Serial.println("Start");

  rollers(ROLLERS_ON);
  delay(200); 
  digitalWrite(enableMotor,HIGH);   //enable motors
  escape_timer = 60;
  distance_sensor_timer = 5;    //50ms
  
}


//------------------------------------------------------------------------------

//Execute A/D converter and Doppler radar DSP code at fs rate (4800 Hz)
//Determine if opponent is present and left or right
//Execute timer code also

void procSample()             //Interrupt executes this 4800/sec  (frequency set by fs variable) 
{
 
digitalWrite(TEST,HIGH);

 // ****   Doppler radar processing  ********
 
adc->startSingleRead(rDoppler,ADC_0);  //Read right doppler
while(!adc->isComplete(ADC_0));
if(adc->isComplete(ADC_0)){
       rOpponent = adc->readSingle(ADC_0) ;
      
}

adc->startSingleRead(lDoppler,ADC_0);   //Read left doppler
while(!adc->isComplete(ADC_0));
if(adc->isComplete(ADC_0)){
      lOpponent = adc->readSingle(ADC_0) ;
      
}

adc->startSingleRead(POT,ADC_0);      //Read sensitivity pot
while(!adc->isComplete(ADC_0));
if(adc->isComplete(ADC_0)){
      pot = adc->readSingle(ADC_0) ;
      pot = pot / 8;  //range now 0->512
}

  lOpponent -=  2048;                           //Remove offset to make it bipolar
  lOpponent -= lpL1.iterate(lOpponent);         // 0hz DC blocker.  Subtracted lowpass sample from unfiltered sample
  lDC = abs(lOpponent * 32);                    //rectify  left signal to DC
  lDC = lpL2.iterate(lDC);                      //low pass filter left signal to remove ripple
  lDC = lDC/8;                                  //reduce range 

 
  rOpponent -= 2048;                            //Remove offset to make it bipolar
  rOpponent -= lpR1.iterate(rOpponent);        // Ohz Dc blocker.   Subtracted lowpass sample from unfiltered sample
  rDC = abs(rOpponent * 32);                   // rectify right signal to DC
  rDC = lpR2.iterate(rDC);                     // Low pass right signal to remove ripple
  rDC = rDC / 8;                              // reduce range 

  analogWrite(DAC,(rOpponent/8)+128);         //Write sample to DAC for debuging

   if((lDC < pot) && (rDC < pot)) opponentLocation = OD_NONE;
   if((lDC < pot) && (rDC >= pot)) opponentLocation = OD_RIGHT;
   if((lDC >= pot) && (rDC < pot)) opponentLocation = OD_LEFT;
   if((lDC >= pot) && (rDC >= pot)) opponentLocation = OD_CENTER;
   if(distance_mm <= 120) opponentLocation = OD_CENTER_NEAR;    //time of flight distance sensor sees opponent at less than 120mm

  if(opponentLocation != OD_NONE) digitalWrite(LED,HIGH);  else digitalWrite(LED,LOW);

  int32_t whiteLine = digitalRead(lLine) * 1000;      //read whiteline sensor.  It will be zero or 1000
  lpLeftLine.iterate(whiteLine);                      //Low pass filter whiteline
  leftLineFiltered = lpLeftLine.getLP();              //Put filtered data into leftLineFiltered

  whiteLine = digitalRead(rLine) * 1000;              //Again for the right white line sensor
  rightLineFiltered = lpRightLine.iterate(whiteLine);

  if(rightLineFiltered > 500) rightBorder = 1; else rightBorder = 0;  //Right white line sensor
  if(leftLineFiltered > 500) leftBorder = 1; else leftBorder = 0;     //Left white line sensor

  aim = rOpponent; 
  analogWrite(DAC,(rOpponent/4)+512);         //Write sample to DAC for debuging
  

  // ****  Timers   ******

  if(prescale-- == 0){          //Divide fs  down to 100Hz
         
         prescale = fs/100;         // 4800/100=48      
         if(timer != 0) timer--;    //10 ms pertick timer
         if(timer2 != 0) timer2--;
         if(motor_timer != 0) motor_timer--;
         if(m_timer != 0) m_timer--;
         if(distance_sensor_timer != 0) distance_sensor_timer--;
         
      
                 
         if(divide10++ >= 9){
             
             if(startup_timer != 0)    startup_timer--; //100ms per tick timers
             if(escape_timer != 0)     escape_timer--;
             if(cruse_memory != 0)     cruse_memory--;
             if(pushback_memory != 0)  pushback_memory--;
             if(border_memory != 0 )   border_memory--;
             if(tilt_memory != 0 )     tilt_memory--;
             
             divide10 = 0;
             
                                       
         }

         if(divide100++ >= 99){  // 1 second per tick timer
            runTimer++;            
            divide100 = 0;

            
         }

      }
  
  
  digitalWrite(TEST,LOW);
  
}
//--------------------------------------------------------------------------

//Control foam paint rollers on/reverse/off

void rollers(roller cmd)
{
  switch(cmd){
    case ROLLERS_OFF: digitalWrite(rollerPWM,LOW);
              break;
    case ROLLERS_ON: digitalWrite(rollerPWM,HIGH);
                     digitalWrite(rollerDIR,HIGH);
              break;

    case ROLLERS_REV: digitalWrite(rollerPWM,HIGH);
                      digitalWrite(rollerDIR,LOW);
                      break;
                      
  }
  
}

//----------------------------------------------------------------------
void updateMotors()
{
  analogWrite(rMotor,r_speed+127);
  analogWrite(lMotor,l_speed+127);
}

//-----------------------------------------------------------------------

 const int FWD5 = 127;
 const int FWD4 = 100;
 const int FWD3 = 80;
 const int FWD2 = 63;
 const int FWD1 = 40;
 const int FWD0 = 29;
 const int REV0 = -FWD0;
 const int REV1 = -FWD1;
 const int REV2 = -FWD2;
 const int REV3 = -FWD3;
 const int REV4 = -FWD4;
 const int REV5 = -FWD5;
 
   
   
   void dir_control(char value)
   {
      
      

       switch(value){

       
       case FWD:     l_speed = FWD4;
                     r_speed = FWD4;
                     break;

       case FWD_MAX: l_speed = FWD5;
                     r_speed = FWD5;
                     break;

       case FWD_LEFT:
                     l_speed = FWD1;
                     r_speed = FWD5;
                     break;

       case FWD_RIGHT:
                     l_speed = FWD5;
                     r_speed = FWD1;
                     break;

       case FWD_HARD_LEFT:
                     l_speed = 0;
                     r_speed = FWD5;
                     break;

       case FWD_HARD_RIGHT:
                     l_speed = FWD5;
                     r_speed = 0;
                     break;

       case REV:     l_speed = REV4;
                     r_speed = REV4;
                     break;

       case REV_MAX: l_speed = REV5;
                     r_speed = REV5;
                     break;


       case REV_LEFT:
                     l_speed = FWD1;
                     r_speed = REV5;
                     break;

       case REV_RIGHT:
                     l_speed = REV5;
                     r_speed = FWD1;
                     break;

       case REV_HARD_LEFT:
                     l_speed = 0;
                     r_speed = REV5;
                     break;

       case REV_HARD_RIGHT:
                     l_speed = REV5;
                     r_speed = 0;
                     break;

       case SPIN_LEFT:
                     l_speed = REV3;
                     r_speed = FWD3;
                     break;

       case SPIN_RIGHT:
                     l_speed = FWD3;
                     r_speed = REV3;
                     break;

       case STOP:
      
                     l_speed = 0;
                     r_speed = 0;
                     break;

       case BRAKE:   l_speed = REV0;     
                     r_speed = REV0;
                     break;

       }

       updateMotors();
   }


   
  //---------------------------------------------------------------------------------

   #define P_NORMAL 5
   #define P_MAX 255

      
   #define M_BACKUP  40
   #define M_TURNAROUND  41
   #define M_BACKUP_TURNAROUND  42
   #define M_ESCAPE_L 43
   #define M_ESCAPE_R 44
   #define M_BRAKE 45
   #define M_TIMED_BACKUP 46
   #define M_ESCAPE_REV 47
   
   void maneuver()
   {
      static char m_sens;

      switch(m_action){

      case M_TIMED_BACKUP:

         maneuver_output = REV;
         break;


      case M_BACKUP_TURNAROUND:
      
      m_sens = 0;
      if(leftBorder) { bitSet(m_sens,0); } //Remember which sensor saw border line
      if(rightBorder){ bitSet(m_sens,1); }
      rollers(ROLLERS_REV);
      m_action = M_BACKUP;
      //Fall through to M_BACKUP
    
      
      case M_BACKUP: 

         maneuver_output = REV;

         if((!leftBorder) 
            && (!rightBorder)
            && (m_border)
            && (m_timer < 30))
              { m_action = M_TURNAROUND;
                m_timer = 50;         //A little more than 180 degrees
              }

         if(m_timer == 0){
            if((leftBorder) && (rightBorder)){  //We are blocked from the rear
               maneuver_output = REV_MAX;                  //with both sensors on the border line.
               m_timer = 50;
            }

            if((leftBorder) && (!rightBorder)){ //Blocked from rear with left sensor on line.
                 m_timer = 30;
                 m_action = M_ESCAPE_R;

            }

           
            if((!leftBorder) && (rightBorder)){ //Blocked from rear with right sensor on line.
               m_timer = 30;
               m_action = M_ESCAPE_L;

            }
         }

         break;
      
      case M_ESCAPE_R:
            maneuver_output =SPIN_RIGHT;
            if(leftBorder){
               m_timer = 30;          //Continue to rotate right 300ms after we are off the border line.
            }


         break;

      case M_ESCAPE_L:
            maneuver_output = SPIN_LEFT;
            if(rightBorder){
               m_timer = 30;        //Continue to rotate left 300ms after we are off the border line.
            }


         break;

       case M_ESCAPE_REV:
            maneuver_output = REV;         //back up briefly to try to break deadlock   
            break;

      case M_TURNAROUND:

                    
           if(m_sens == 1) { 
              maneuver_output = SPIN_RIGHT ;
              rollers(ROLLERS_REV);
              
            }

           if(m_sens & 2) { 
              maneuver_output = SPIN_LEFT ; 
              rollers(ROLLERS_REV);
           }

           if((m_timer < 25) && (opponentLocation == OD_NONE)){ m_timer = 0;} //Stop turning if we see opponent

           if(m_timer == 0){
                  
                  m_action = M_BRAKE;
                  m_timer = 5;
               }

                     

         break;

         

         case M_BRAKE:
               
               maneuver_output =  BRAKE;
               break;
         
        
            
      }






      if((m_timer == 0) || (m_action == 0)) {
         m_action = 0;           //Catch all. Limit timer terminates is default.
         m_timer = 0;
         m_priority = 0;
         maneuver_output = 0;   
         rollers(ROLLERS_ON);      
         
      }
 


   }

//---------------------------------------------------------------------------

void cancel_maneuver()
{
   m_action = 0;
   m_timer = 0;
   maneuver_output = 0;
   
   

}

//---------------------------------------------------------------------------
//  Call this function to start a maneuver.  Maneuvers in progress can only
//  be interrupted by a higher priority maneuver.
//  Args: MANEUVER_CODE, time limit 10ms ticks, border, opponent, PRIORITY
//  Caller sets the maneuver action and termination conditions in args above.
void start_maneuver(char a, char t, char b, char o, char p)
{
   
   
   if((m_action != 0) && (p > m_priority)){
      cancel_maneuver();
   }

   if(m_action == 0){
      m_action = a;
      m_timer = t;     // 10ms per tick maneuver limit timer (1..255 = 10ms to 2.55 sec. )
      m_border = b;
      m_opponent = o;
      m_priority = p;
      
      

   }


}

//---------------------------------------------------------------------------

void check_border()
{
 if(leftBorder || rightBorder){

    escape_timer = 60;      // escape_timer will time out at  in 6 seconds if no border line seen again
    start_maneuver(M_BACKUP_TURNAROUND,70,1,1,P_MAX);
    
 }

}

 
 //------------------------------------------------------------------------
  
void cruse()
   {  
      cruse_output = FWD;
     
          
   }
   
 //-----------------------------------------------------------------------

 


 
 void chase()
   {  
      
      chase_output = 0;
      
      bitClear(flags,attack);
     

      switch(opponentLocation){
      case OD_LEFT:  chase_output = FWD_LEFT;  break;
      case OD_RIGHT: chase_output = FWD_RIGHT; break;
      case OD_CENTER: chase_output = FWD; break;
      case OD_CENTER_NEAR: chase_output = FWD_MAX; bitSet(flags,attack); break;
      default:    chase_output = 0; break;
      }

     
      
      
   }
   //-----------------------------------------------------------------------

 void break_deadlock()
 {
    
    if(escape_timer == 0){                            //6 seconds of no border white line crossings means deadlock
      escape_timer = 60;                              //Reset timer to 6 seconds
      start_maneuver(M_ESCAPE_REV,10,0,0,P_NORMAL);   //100ms reverse to break deadlock
      
    }  
 }

 //--------------------------------------------------------------------------

   // Subsumption arbitration happens here.
   void arbitrate(){

      
     arb_code = 0;
     
     if(cruse_output) { arb_code = 1; dir_ctl = cruse_output; }                 //Lowest priority
     if(chase_output) { arb_code = 2; dir_ctl = chase_output; }         
     if(maneuver_output) {arb_code = 3; dir_ctl = maneuver_output;}           //Highest priority
     
     dir_control(dir_ctl);   //Now tell the motors what to do.
     
   }

   //-------------------------------------------------------------------------

void loop() {

//rollers(ROLLERS_OFF);
//digitalWrite(enableMotor,LOW); 

if(millis() > ptime){
ptime = millis() + 200;
//Serial << lDC << " " << rDC << " " << pot << " " << distance_mm << " " << l_speed<< " " << r_speed<< endl ;   //Debug

}

if(distance_sensor_timer == 0){
  distance_sensor_timer = 5;  //50ms
  distance_mm = distance_sensor.readRangeContinuousMillimeters();
}

if(opponentLocation == OD_CENTER_NEAR) digitalWrite(LED2,HIGH); else digitalWrite(LED2,LOW);


check_border();
maneuver();
cruse();
chase();
break_deadlock();
arbitrate();



}
